package uk.ac.starlink.auth;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;

/**
 * IvoaAuthScheme concrete subclass for working with bearer tokens.
 *
 * <p>This implements the {@value #SCHEME_NAME} scheme.
 * The bearer token is acquired from the {@value #TOKEN_HEADER} header
 * of the login response, and is subsequently used in accordance with
 * RFC6750 like:
 * <pre>
 *    Authorization: Bearer &lt;token-text&gt;
 * </pre>
 *
 * @author   Mark Taylor
 * @since    10 Dec 2021
 * @see    <a href="https://datatracker.ietf.org/doc/html/rfc6750">RFC 6750</a>
 */
public class BearerIvoaAuthScheme extends IvoaAuthScheme {

    /** Singleton instance. */
    public static final BearerIvoaAuthScheme INSTANCE =
        new BearerIvoaAuthScheme();

    /** Name and identifier for this scheme. */
    public static final String SCHEME_NAME = "ivoa_bearer";

    /** Name of login response header containing bearer token. */
    public static final String TOKEN_HEADER = "X-VO-Bearer";

    /**
     * Private constructor prevents instantiation.
     */
    private BearerIvoaAuthScheme() {
        super( SCHEME_NAME );
    }

    public ContextFactory createContextFactory( LoginProtocol loginProto,
                                                URL loginUrl,
                                                URL challengeUrl ) {
        final ProtectionSpace scope = new ProtectionSpace( challengeUrl, null );
        return new ContextFactory() {
            public AuthContext createContext( UserInterface ui ) {
                String token = readAuth( loginProto, loginUrl, ui,
                                         BearerIvoaAuthScheme::readBearerToken);
                return createBearerContext( token );
            }
            public AuthContext createUnauthContext() {
                return createBearerContext( null );
            }
            private AuthContext createBearerContext( String token ) {
                return new BearerContext( BearerIvoaAuthScheme.this,
                                          loginUrl.toString(), scope, token );
            }
        };
    }

    /**
     * Reads a bearer token from a login response header.
     *
     * @param  conn  login response connection
     * @return   bearer token, not null
     */
    private static String readBearerToken( HttpURLConnection conn )
            throws IOException {
        String token = conn.getHeaderField( TOKEN_HEADER );
        conn.getInputStream().close();
        if ( token == null || token.trim().length() == 0 ) {
            throw new IOException( "No " + TOKEN_HEADER
                                 + " field in login response" );
        }
        else {
            return token;
        }
    }

    /**
     * AuthContext implementation for Bearer tokens.
     *
     * <p>Note this uses the context of a validity scope, which is not
     * currently defined by the ivoa_bearer scheme.
     * But the assumption is made that if this factory is suitable for
     * one URL in a given ProtectionScheme (host+port),
     * it will be suitable for all others.
     */
    private static class BearerContext implements AuthContext {

        private final BearerIvoaAuthScheme scheme_;
        private final String loginUrl_;
        private final ProtectionSpace scope_;
        private final String token_;

        /**
         * Constructor.
         *
         * @param  scheme  scheme 
         * @param  loginUrl   URL at which the token can be acquired
         * @param  scope   URL range for which this context is valid
         * @param  token   IVOA bearer token, or null for anonymous
         */
        BearerContext( BearerIvoaAuthScheme scheme, String loginUrl,
                       ProtectionSpace scope, String token ) {
            scheme_ = scheme;
            loginUrl_ = loginUrl;
            scope_ = scope;
            token_ = token;
        }

        public AuthScheme getScheme() {
            return scheme_;
        }

        public boolean hasCredentials() {
            return token_ != null;
        }

        public boolean isUrlDomain( URL url ) {
            return scope_.equals( new ProtectionSpace( url, null ) );
        }

        public boolean isChallengeDomain( Challenge challenge, URL url ) {
            return scope_.equals( new ProtectionSpace( url, null ) )
                && loginUrl_
                  .equals( challenge.getParams().get( ACCESSURL_PARAM ) );
        }

        public boolean isExpired() {
            return false;
        }

        public void configureConnection( HttpURLConnection connection )
                throws IOException {
            if ( token_ != null ) {
                connection.setRequestProperty( AuthUtil.AUTH_HEADER,
                                               "Bearer " + token_ );
            }
        }

        public String[] getCurlArgs( URL url, boolean showSecrets ) {
            return new String[] {
                "--header",
                AuthUtil.AUTH_HEADER + ": Bearer " + token_,
            };
        }
    }
}
