package uk.ac.starlink.auth;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * AuthScheme implementation that works (at time of writing)
 * with the implementation of the CADC services.
 *
 * <p>This scheme looks for a challenge like:
 * <pre>
 *    www-authenticate: ivoa_bearer standard_id="ivo://ivoa.net/sso#tls-with-password", access_url="https://ws-cadc.canfar.net/ac/login"
 * </pre>
 * and the text/plain content of the resource at the given
 * <code>access_url</code>, when authenticated <code>username</code> and
 * <code>password</code> parameters are supplied, is a Bearer token that
 * can be used in accordance with RFC6750:
 * <pre>
 *    Authorization: Bearer &lt;token-text&gt;
 * </pre>
 *
 * @author  Mark Taylor
 * @since   20 May 2021
 * @see  <a href="https://ws-cadc.canfar.net/youcat"
 *               >https://ws-cadc.canfar.net/youcat</a>
 */
public class CadcAuthScheme implements AuthScheme {

    /** Singleton instance. */
    public static final CadcAuthScheme INSTANCE = new CadcAuthScheme();

    private static final String SSO_URI = "ivo://ivoa.net/sso";
    private static final String TLS_PASSWD_ID = SSO_URI + "#tls-with-password";
    private static final String STANDARDID_PARAM = "standard_id";
    private static final String ACCESSURL_PARAM = "access_url";
    private static final String USERNAME = "username";
    private static final String PASSWORD = "password";
    private static final int MAX_TOKEN_LENGTH = 16 * 1024;

    private CadcAuthScheme() {
    }

    public String getName() {
        return "CADC";
    }

    public ContextFactory createContextFactory( Challenge challenge, URL url )
            throws BadChallengeException {
        if ( "ivoa_bearer".equalsIgnoreCase( challenge.getSchemeName() ) ) {
            String stdId = challenge.getRequiredParameter( STANDARDID_PARAM );
            ProtectionSpace scope = new ProtectionSpace( url, null );
            if ( stdId.equalsIgnoreCase( TLS_PASSWD_ID ) ) {
                URL loginUrl =
                    challenge.getRequiredParameterUrl( ACCESSURL_PARAM );
                return new CadcContextFactory( this, loginUrl, scope );
            }
            else {
                return null;
            }
        }
        else {
            return null;
        }
    }

    /**
     * Acquires a bearer token from the given URL, with credential input
     * from the user.
     *
     * @param  loginUrl  URL at which credentials can be swapped for a token
     * @param  ui   user interface for supplying credentials
     */
    private String readLoginToken( URL loginUrl, UserInterface ui )
            throws IOException {
        String msg = "IVOA login to " + loginUrl;
        do {
            UserPass userpass = ui.readUserPassword( msg );
            if ( userpass == null ) {
                return null;
            }
            Map<String,String> postParams = new LinkedHashMap<String,String>();
            postParams.put( USERNAME, userpass.getUsername() );
            postParams.put( PASSWORD, new String( userpass.getPassword() ) );
            HttpURLConnection conn =
                (HttpURLConnection) loginUrl.openConnection();
            conn.setRequestProperty( "Accept", "text/plain" );
            AuthUtil.postForm( conn, postParams );
            int code = conn.getResponseCode();
            if ( code == 200 ) {

                /* May throw an IOException, but if it does retrying
                 * isn't much likely to help, so don't bother to catch it. */
                return readToken( conn );
            }
            else if ( code == 401 || code == 403 ) {
                String failMsg = AuthUtil.authFailureMessage( conn );
                if ( ui.canRetry() ) {
                    ui.message( failMsg );
                }
                else {
                    throw new IOException( "Login attempt failed: " + failMsg );
                }
            }
            else {
                String failMsg = AuthUtil.authFailureMessage( conn );
                throw new IOException( "Login error: " + failMsg );
            }
        } while ( ui.canRetry() );
        return null;
    }

    /**
     * Reads the plain text content from a given connection.
     *
     * @throws  IOException if the content is too long,
     *                      or the content-type is not text/plain,
     *                      or something else went wrong
     */
    private String readToken( HttpURLConnection conn ) throws IOException {
        String mimeType = conn.getContentType();
        ContentType ctype = ContentType.parse( mimeType );
        if ( ctype != null && ! ctype.matches( "text", "plain" ) ) {
            throw new IOException( "Unexpected Content-Type " + mimeType );
        }
        StringBuffer sbuf = new StringBuffer();
        InputStream in = new BufferedInputStream( conn.getInputStream() );
        try {
            for ( int b; ( b = in.read() ) >= 0; ) {
                sbuf.append( (char) b );

                /* If for some reason the content is much longer than the
                 * plausible length for a token, bail out. */
                if ( sbuf.length() > MAX_TOKEN_LENGTH ) {
                    throw new IOException( "Token length overflow" );
                }
            }
            return sbuf.toString();
        }
        finally {
            in.close();
        }
    }

    /**
     * ContextFactory implementation for CADC scheme.
     */
    private static class CadcContextFactory implements ContextFactory {

        private final CadcAuthScheme scheme_;
        private final URL loginUrl_;
        private final ProtectionSpace scope_;

        /**
         * Constructor.
         *
         * @param  scheme  scheme
         * @param  loginUrl   URL at which the token can be acquired
         * @param  scope   URL range for which this context is valid
         */
        CadcContextFactory( CadcAuthScheme scheme, URL loginUrl,
                            ProtectionSpace scope ) {
            scheme_ = scheme;
            loginUrl_ = loginUrl;
            scope_ = scope;
        }

        public AuthContext createContext( UserInterface ui ) {
            final String token;
            try {
                token = scheme_.readLoginToken( loginUrl_, ui );
            }
            catch ( IOException e ) {
                ui.message( "TLS login failed: " + e );
                return null;
            }
            return token == null
                 ? null
                 : new CadcContext( scheme_, loginUrl_.toString(), scope_,
                                    token );
        }

        public AuthContext createUnauthContext() {
            return new CadcContext( scheme_, loginUrl_.toString(), scope_,
                                    null );
        }
    }

    /**
     * AuthContext implementation for CADC.
     *
     * <p>Note this uses the context of a validity scope, which is not
     * defined by the CADC scheme.  But the assumption is made that
     * if this factory is suitable for one URL in a given ProtectionScheme
     * (host+port), it will be suitable for all others.
     */
    private static class CadcContext implements AuthContext {

        private final CadcAuthScheme scheme_;
        private final String loginUrl_;
        private final ProtectionSpace scope_;
        private final String token_;

        /**
         * Constructor.
         *
         * @param  scheme  scheme
         * @param  loginUrl   URL at which the token can be acquired
         * @param  scope   URL range for which this context is valid
         * @param  token   IVOA bearer token, or null for anonymous
         */
        CadcContext( CadcAuthScheme scheme, String loginUrl,
                     ProtectionSpace scope, String token ) {
            scheme_ = scheme;
            loginUrl_ = loginUrl;
            scope_ = scope;
            token_ = token;
        }

        public AuthScheme getScheme() {
            return scheme_;
        }

        public boolean hasCredentials() {
            return token_ != null;
        }

        public boolean isUrlDomain( URL url ) {
            return scope_.equals( new ProtectionSpace( url, null ) );
        }

        public boolean isChallengeDomain( Challenge challenge, URL url ) {
            return scope_.equals( new ProtectionSpace( url, null ) )
                && loginUrl_
                  .equals( challenge.getParams().get( ACCESSURL_PARAM ) );
        }

        public boolean isExpired() {
            return false;
        }

        public void configureConnection( HttpURLConnection connection )
                throws IOException {
            if ( token_ != null ) {
                connection.setRequestProperty( AuthUtil.AUTH_HEADER,
                                               "Bearer " + token_ );
            }
        }

        public String[] getCurlArgs( URL url, boolean showSecrets ) {
            return new String[] {
                "--header",
                AuthUtil.AUTH_HEADER + ": Bearer " + token_,
            };
        }
    }
}
