package uk.ac.starlink.auth;

import java.io.IOException;
import java.net.CookieManager;
import java.net.CookieStore;
import java.net.HttpCookie;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

/**
 * AuthScheme implementation that works with the cookie scheme as
 * prototyped by the GACS system.  Their example challenge looks
 * like this:
 * <pre>
 *   WWW-Authenticate: ivo://ivoa.net/std/SSO#cookie
 *                     realm="Authentication needed",
 *                     accessURL="https://geadev.esac.esa.int/tap-server/login",
 *                     userParameter="username", passwordParameter="password",
 *                     responseCookie="JSESSIONID"
 * </pre>
 *
 * <p>This is inferior in a few ways to the GenericCookieAuthScheme;
 * the responseCookie name and realm are not really necessary.
 * And it doesn't use a CookieManager.  Just for comparison really.
 *
 * @author   Mark Taylor
 * @since    15 Jun 2020
 * @see  <a href="https://tools.ietf.org/html/rfc2965">RFC 2965</a>
 */
public class GacsCookieAuthScheme implements AuthScheme {

    public static final GacsCookieAuthScheme INSTANCE =
        new GacsCookieAuthScheme();

    private static final String ACCESSURL_PARAM = "accessurl";
    private static final String RESPONSECOOKIE_PARAM = "responsecookie";
    private static final String USERPARAMETER_PARAM = "userparameter";
    private static final String PASSWORDPARAMETER_PARAM = "passwordparameter";

    private static final Logger logger_ =
        Logger.getLogger( "uk.ac.starlink.auth" );

    private GacsCookieAuthScheme() {
    }

    public String getName() {
        return "cookie-gacs";
    }

    public ContextFactory createContextFactory( Challenge challenge, URL url )
            throws BadChallengeException {
        if ( "ivo://ivoa.net/std/SSO#cookie"
            .equalsIgnoreCase( challenge.getSchemeName() ) ) {
            URL cookieUrl =
                challenge.getRequiredParameterUrl( ACCESSURL_PARAM );
            String cookieName =
                challenge.getRequiredParameter( RESPONSECOOKIE_PARAM );
            String userKey =
                challenge.getRequiredParameter( USERPARAMETER_PARAM );
            String passwordKey =
                challenge.getRequiredParameter( PASSWORDPARAMETER_PARAM );
            return new GacsCookieContextFactory( this, challenge.getRealm(),
                                                 cookieUrl, cookieName,
                                                 userKey, passwordKey );
        }
        else {
            return null;
        }
    }
    
    private HttpCookie readCookie( URL cookieUrl, String cookieName,
                                   String userKey, String passwordKey,
                                   UserInterface ui )
            throws IOException {

        /* Privacy implications for starting a cookie-based session:
         * see RFC2965 sec 6.1. */
        String msg = "Cookie-based authentication to "
                   + cookieUrl.getAuthority();
        do {
            UserPass userpass = ui.readUserPassword( msg );
            if ( userpass == null ) {
                return null;
            }
            Map<String,String> postParams = new LinkedHashMap<String,String>();
            postParams.put( userKey, userpass.getUsername() );
            postParams.put( passwordKey, new String( userpass.getPassword() ) );
            HttpURLConnection conn = AuthUtil.postForm( cookieUrl, postParams );
            HttpCookie[] cookies = readCookies( conn );
            for ( HttpCookie cookie : cookies ) {
                if ( cookieName.equals( cookie.getName() ) &&
                     ! cookie.hasExpired() ) {
                    logger_.info( "Acquired cookie: "
                                + AuthUtil.cookieLogText( cookie ) );
                    return cookie;
                }
            }
        } while ( ui.canRetry() );
        return null;
    }

    private HttpCookie[] readCookies( HttpURLConnection conn ) {
        Map<String,List<String>> headers = conn.getHeaderFields();
        List<HttpCookie> cookies = new ArrayList<HttpCookie>();
        for ( Map.Entry<String,List<String>> entry : headers.entrySet() ) {
            String key = entry.getKey();
            if ( key != null &&
                 ( key.equalsIgnoreCase( "set-cookie" ) ||
                   key.equalsIgnoreCase( "set-cookie2" ) ) ) {
                for ( String cookieTxt : entry.getValue() ) {
                    cookies.addAll( HttpCookie.parse( cookieTxt ) );
                }
            }
        }
        return cookies.toArray( new HttpCookie[ 0 ] );
    }

    private static class GacsCookieContextFactory implements ContextFactory {

        private final GacsCookieAuthScheme scheme_;
        private final String realm_;
        private final URL cookieUrl_;
        private final String cookieName_;
        private final String userKey_;
        private final String passwordKey_;

        GacsCookieContextFactory( GacsCookieAuthScheme scheme, String realm,
                                  URL cookieUrl,
                                  String cookieName, String userKey,
                                  String passwordKey ) {
            scheme_ = scheme;
            realm_ = realm;
            cookieUrl_ = cookieUrl;
            cookieName_ = cookieName;
            userKey_ = userKey;
            passwordKey_ = passwordKey;
        }

        public AuthContext createContext( UserInterface ui ) {
            final HttpCookie cookie;
            try {
                cookie = scheme_.readCookie( cookieUrl_, cookieName_,
                                             userKey_, passwordKey_, ui );
            }
            catch ( IOException e ) {
                ui.message( "Cookie acquisition failed: " + e );
                return null;
            }

            // GDPR: do I need to warn the user they are using a cookie here?
            return cookie == null
                 ? null
                 : new GacsCookieContext( scheme_, cookieUrl_, realm_, cookie );
        }

        public AuthContext createUnauthContext() {
            return new GacsCookieContext( scheme_, cookieUrl_, realm_, null );
        }
    }

    private static class GacsCookieContext implements AuthContext {

        private final GacsCookieAuthScheme scheme_;
        private final String accessurl_;
        private final String realm_;
        private final ProtectionSpace loginScope_;
        private final HttpCookie cookie_;
        private final CookieStore cookieStore_;

        GacsCookieContext( GacsCookieAuthScheme scheme, URL cookieUrl,
                           String realm, HttpCookie cookie ) {
            scheme_ = scheme;
            accessurl_ = cookieUrl.toString();
            realm_ = realm;
            loginScope_ = new ProtectionSpace( cookieUrl, null );
            cookie_ = cookie;
            if ( cookie != null ) {
                cookieStore_ = new CookieManager().getCookieStore();
                cookieStore_.add( (URI) null, cookie );
            }
            else {
                cookieStore_ = null;
            }
        }

        public AuthScheme getScheme() {
            return scheme_;
        }

        public boolean hasCredentials() {
            return cookie_ != null;
        }

        public void configureConnection( HttpURLConnection connection ) {
            // worry about unverifiable transactions here?  RFC2965 sec 3.3.6.
            if ( cookie_ != null ) {
                connection.setRequestProperty( "Cookie", cookie_.toString() );
            }
        }

        public String[] getCurlArgs( URL url, boolean showSecret ) {
            if ( cookie_ == null ) {
                return new String[ 0 ];
            }
            else {
                return new String[] {
                    "--header",
                    "Cookie: " + ( showSecret ? cookie_.toString()
                                              : ( cookie_.getName()
                                                + "=<cookie-value>" ) ),
                };
            }
        }

        public boolean isUrlDomain( URL url ) {
            if ( cookie_ == null ) {
                return false;
            }

            // RFC 2965 sec 3.3.4.
            URI uri;
            try {
                uri = url.toURI();
            }
            catch ( URISyntaxException e ) {
                return false;
            }
            List<HttpCookie> applicables = cookieStore_.get( uri );
            if ( applicables.size() == 0 ) {
                return false;
            }
            else if ( applicables.size() == 1 ) {
                assert applicables.get( 0 ) == cookie_;
                return true;
            }
            else {
                assert false;
                return false;
            }
        }

        public boolean isChallengeDomain( Challenge challenge, URL url ) {
            try {
                return scheme_.createContextFactory( challenge, url ) != null
                    && AuthUtil.unNullString( realm_ )
                      .equals( AuthUtil.unNullString( challenge.getRealm() ) )
                    && accessurl_.equals( challenge.getParams()
                                                   .get( ACCESSURL_PARAM ) )
                    && loginScope_.equals( new ProtectionSpace( url, null ) );
            }
            catch ( BadChallengeException e ) {
                return false;
            }
        }

        public boolean isExpired() {
            return cookie_ != null && cookie_.hasExpired();
        }
    }
}
