package uk.ac.starlink.auth;

import java.io.IOException;
import java.net.CookieManager;
import java.net.CookieStore;
import java.net.HttpCookie;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

/**
 * Implements an authentication scheme that requires a cookie
 * (or possibly cookies) for authentication, and in which the
 * source of the cookie is given in the challenge.
 * This is not an internet standard.
 *
 * <p>A challenge looks like this:
 * <pre>   WWW-Authenticate: /ivoa.net/std/SSO#cookie accessURL=&lt;loginurl&gt;
 * </pre>
 * The client then POSTs a form to the URL given by the
 * <code>{@value #ACCESSURL_PARAM}</code> parameter,
 * with suitable values for the
 * <code>{@value #USERNAME}</code> and <code>{@value #PASSWORD}</code> keys,
 * and remembers the cookie (or possibly cookies) it gets in response.
 * That cookie can then be used for access to the protected resources.
 * The cookie itself contains relevant scope information.
 * A <code>realm</code> parameter if supplied is respected for
 * assessing other challenges, but it's not usually required because of
 * the scoping provided by the cookie.
 *
 * @author   Mark Taylor
 * @since    15 Jun 2020
 * @see  <a href="https://tools.ietf.org/html/rfc2965">RFC 2965</a>
 */
public class GenericCookieAuthScheme implements AuthScheme {

    /** Singleton instance. */
    public static final GenericCookieAuthScheme INSTANCE =
        new GenericCookieAuthScheme();

    /** Name of challenge parameter giving login URL. */
    public static final String ACCESSURL_PARAM = "accessurl";

    /** Key for POSTing username to login URL. */
    public static final String USERNAME = "username";

    /** Key for POSTing password to login URL. */
    public static final String PASSWORD = "password";

    private static final Logger logger_ =
        Logger.getLogger( "uk.ac.starlink.auth" );

    /**
     * Private constructor prevents external instantiation of singleton class.
     */
    private GenericCookieAuthScheme() {
    }

    public String getName() {
        return "cookie-generic";
    }

    public ContextFactory createContextFactory( Challenge challenge, URL url )
            throws BadChallengeException {
        if ( "ivo://ivoa.net/std/SSO#cookie"
            .equalsIgnoreCase( challenge.getSchemeName() ) ) {
            String realm = challenge.getRealm();
            URL cookieUrl = challenge.getRequiredParameterUrl( ACCESSURL_PARAM);
            return new GenericCookieContextFactory( this, cookieUrl, realm );
        }
        else {
            return null;
        }
    }

    /**
     * Attempt to log in to a given URL and return the cookies that result
     * from doing so.
     * The return value is a CookieManager rather than a simple HttpCookie,
     * since the CookieManager knows how to do all the validity calculus
     * defined in RFC2965 sec 3.3.4.
     *
     * @param   cookieUrl   login URL
     * @param   ui   user interface abstraction
     * @return   cookie manager holding returned cookies,
     *           or null if no cookies were acquired
     */
    private CookieManager readCookies( URL cookieUrl, UserInterface ui )
            throws IOException {

        /* There are privacy implications for starting a cookie-based
         * session (see RFC2965 sec 6.1), so it's good practice to inform
         * the user that this is going on.  However, given that the whole
         * point of this is authentication, the privacy implications
         * (that the server could "track the path of a user through the
         * server") ought to be pretty obvious. */
        String msg = "Cookie-based authentication to "
                   + cookieUrl.getAuthority();
        do {
            UserPass userpass = ui.readUserPassword( msg );
            if ( userpass == null ) {
                return null;
            }
            Map<String,String> postParams = new LinkedHashMap<String,String>();
            postParams.put( USERNAME, userpass.getUsername() );
            postParams.put( PASSWORD, new String( userpass.getPassword() ) );
            HttpURLConnection conn = AuthUtil.postForm( cookieUrl, postParams );
            int code = conn.getResponseCode();
            if ( code == 401 || code == 403 ) {
                String failMsg = AuthUtil.authFailureMessage( conn );
                if ( ui.canRetry() ) {
                    ui.message( failMsg );
                }
                else {
                    throw new IOException( "Cookie authentication failed: "
                                         + failMsg );
                }
            }
            else {
                CookieManager cookieManager = readCookies( conn );
                conn.getInputStream().close();
                List<HttpCookie> cookies =
                    cookieManager.getCookieStore().getCookies();
                int ncookie = cookies.size();
	        if ( ncookie > 0 ) {
                    StringBuffer sbuf = new StringBuffer();
                    if ( ncookie == 1 ) {
                        HttpCookie cookie = cookies.get( 0 );
                        sbuf.append( "Acquired cookie: " )
                            .append( AuthUtil.cookieLogText( cookie ) );
                    }
                    else {
                        sbuf.append( "Acquired cookies:" );
                        for ( HttpCookie cookie : cookies ) {
                            sbuf.append( ' ' )
                                .append( AuthUtil.cookieLogText( cookie ) );
                        }
                    }
                    logger_.info( sbuf.toString() );
                    return cookieManager;
                }
            }
        } while ( ui.canRetry() );
        return null;
    }

    /**
     * Extract returned cookies from an HTTP connection.
     *
     * @param  conn  opened HTTP connection
     * @return  cookie manager containing any cookies present in the response
     *          (possibly none)
     */
    private CookieManager readCookies( HttpURLConnection conn )
            throws IOException {
        CookieManager cookieManager = new CookieManager();
        try {
            URI uri = conn.getURL().toURI();
            cookieManager.put( uri, conn.getHeaderFields() );
        }
        catch ( URISyntaxException e ) {
        }
        return cookieManager;
    }

    /**
     * ContextFactory implementation for use with cookie auth scheme.
     */
    private static class GenericCookieContextFactory implements ContextFactory {

        private final GenericCookieAuthScheme scheme_;
        private final URL cookieUrl_;
        private final String realm_;

        /**
         * Constructor.
         *
         * @param   scheme   auth scheme from which this context arose
         * @param  cookieUrl   cookie access URL
         * @param  realm     challenge realm
         */
        GenericCookieContextFactory( GenericCookieAuthScheme scheme,
                                     URL cookieUrl, String realm ) {
            scheme_ = scheme;
            cookieUrl_ = cookieUrl;
            realm_ = realm;
        }

        public AuthContext createContext( UserInterface ui ) {

            /* Attempt login and acquire resulting cookies. */
            final CookieManager cookieManager;
            try {
                cookieManager = scheme_.readCookies( cookieUrl_, ui );
            }
            catch ( IOException e ) {
                ui.message( "Cookie acquisition failed: " + e );
                return null;
            }

            /* If cookie acquisition was successful, return an AuthContext
             * based on the acquired cookie(s). */
            // GDPR: do I need to warn the user they are using a cookie here?
            return cookieManager == null
                 ? null
                 : new GenericCookieContext( scheme_, cookieUrl_, realm_,
                                             cookieManager );
        }

        public AuthContext createUnauthContext() {
            return new GenericCookieContext( scheme_, cookieUrl_, realm_,
                                             null );
        }
    }


    /**
     * AuthContext implementation used by this class.
     */
    private static class GenericCookieContext implements AuthContext {

        private final GenericCookieAuthScheme scheme_;
        private final String accessurl_;
        private final String realm_;
        private final ProtectionSpace loginSpace_;
        private final CookieManager cookieManager_;

        /**
         * Constructor.
         *
         * @param   scheme   auth scheme from which this context arose
         * @param  cookieUrl   cookie access URL
         * @param  realm     challenge realm
         * @param  cookieManager   cookie manager containing cookies to
         *                         present for authentication,
         *                         or null for unauthenticated access
         */
        GenericCookieContext( GenericCookieAuthScheme scheme,
                              URL cookieUrl, String realm,
                              CookieManager cookieManager ) {
            scheme_ = scheme;
            accessurl_ = cookieUrl.toString();
            realm_ = realm;
            cookieManager_ = cookieManager;
            loginSpace_ = new ProtectionSpace( cookieUrl, null );
        }

        public AuthScheme getScheme() {
            return scheme_;
        }

        public boolean hasCredentials() {
            return cookieManager_ != null;
        }

        public void configureConnection( HttpURLConnection connection )
                throws IOException {
            if ( cookieManager_ != null ) {
                URI uri;
                try {
                    uri = connection.getURL().toURI();
                }
                catch ( URISyntaxException e ) {
                    assert false;
                    return;
                }

                /* Extract and present all the cookies that are relevant for
                 * this connection (the CookieManager takes care of working
                 * out which ones they are). */
                Map<String,List<String>> cookieProps =
                    cookieManager_.get( uri,
                                        connection.getRequestProperties() );
                for ( Map.Entry<String,List<String>> entry :
                      cookieProps.entrySet() ) {
                    String cookieKey = entry.getKey();
                    for ( String cookieValue : entry.getValue() ) {
                        connection.addRequestProperty( cookieKey,
                                                       cookieValue );
                    }
                }
            }
        }

        public String[] getCurlArgs( URL url, boolean showSecret ) {
            if ( cookieManager_ == null ) {
                return new String[ 0 ];
            }
            else {
                URI uri;
                try {
                    uri = url.toURI();
                }
                catch ( URISyntaxException e ) {
                    return new String[ 0 ];
                }
                Map<String,List<String>> cookieProps;
                try {
                    cookieProps =
                        cookieManager_
                       .get( uri, new HashMap<String,List<String>>() );
                }
                catch ( IOException e ) {
                    return new String[ 0 ];
                }
                List<String> args = new ArrayList<String>();
                for ( Map.Entry<String,List<String>> entry :
                      cookieProps.entrySet() ) {
                    String cookieKey = entry.getKey();
                    for ( String cookieValue : entry.getValue() ) {
                        args.add( "--header" );
                        args.add( cookieKey + ": " + cookieValue );
                    }
                }
                return args.toArray( new String[ 0 ] );
            }
        }

        public boolean isUrlDomain( URL url ) {
            URI uri;
            try {
                uri = url.toURI();
            }
            catch ( URISyntaxException e ) {
                return false;
            }

            // RFC 2965 sec 3.3.4.
            return cookieManager_ != null
                && cookieManager_.getCookieStore().get( uri ).size() > 0;
        }

        public boolean isChallengeDomain( Challenge challenge, URL url ) {
            try {
                return scheme_.createContextFactory( challenge, url ) != null
                    && accessurl_.equals( challenge.getParams()
                                                   .get( ACCESSURL_PARAM ) )
                    && AuthUtil.unNullString( realm_ )
                      .equals( AuthUtil.unNullString( challenge.getRealm() ) )
                    && loginSpace_.equals( new ProtectionSpace( url, null ) );
            }
            catch ( BadChallengeException e ) {
                return false;
            }
        }

        public boolean isExpired() {

            /* The CookieManager automatically purges expired cookies. */
            return cookieManager_ != null
                && cookieManager_.getCookieStore().getCookies().size() == 0;
        }
    }
}
